# w.c.s. - web application for online forms
# Copyright (C) 2005-2014  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import json

from quixote import get_publisher
from wcs.roles import Role
from qommon.ctl import Command
from qommon.publisher import get_cfg


class CmdHoboNotify(Command):
    name = 'hobo_notify'

    def execute(self, base_options, sub_options, args):
        self.base_options = base_options
        if sub_options.extra:
            if not self.config.has_section('extra'):
                self.config.add_section('extra')
            for i, extra in enumerate(sub_options.extra):
                self.config.set('extra', 'cmd_line_extra_%d' % i, extra)

        notification = self.load_notification(args)
        if not self.check_valid_notification(notification):
            sys.exit(1)
        import publisher

        publisher.WcsPublisher.configure(self.config)
        pub = publisher.WcsPublisher.create_publisher()
        global_app_dir = pub.app_dir
        for hostname in os.listdir(global_app_dir):
            app_dir = os.path.join(global_app_dir, hostname)
            if not os.path.exists(os.path.join(app_dir, 'config.pck')):
                continue
            pub.app_dir = app_dir
            pub.set_config()
            self.process_notification(notification, pub)

    @classmethod
    def load_notification(cls, args):
        if args[0] == '-':
            # get environment definition from stdin
            return json.load(sys.stdin)
        else:
            return json.load(file(args[0]))

    @classmethod
    def check_valid_notification(cls, notification):
        return isinstance(notification, dict) \
            and '@type' in notification \
            and notification['@type'] in ['provision', 'deprovision'] \
            and 'objects' in notification \
            and 'audience' in notification \
            and isinstance(notification['audience'], list) \
            and isinstance(notification['objects'], list)

    @classmethod
    def process_notification(cls, notification, publisher=None):
        publisher = publisher or get_publisher()
        action = notification['@type']
        audience = notification['audience']
        full = notification['full'] if 'full' in notification else False

        # Verify tenant is in audience
        entity_id = get_cfg('sp', {}).get('saml2_providerid')
        if not entity_id or entity_id not in audience:
            return

        uuids = set()
        # Now provision/deprovision
        for o in notification['objects']:
            t = o['@type']
            if t != 'role' \
               or 'uuid' not in o \
               or 'name' not in o \
               or 'description' not in o \
               or 'emails' not in o \
               or 'emails_to_members' not in o \
               or 'slug' not in o:
                continue
            uuid = o['uuid'].encode(publisher.site_charset)
            uuids.add(uuid)
            slug = o['slug'].encode(publisher.site_charset)
            details = o.get('description', '').encode(publisher.site_charset)
            name = o['name'].encode(publisher.site_charset)
            emails = [email.encode(publisher.site_charset) for email in o['emails']]
            emails_to_members = o['emails_to_members']
            # Find existing role
            try:
                role = Role.get(uuid)
            except KeyError:
                try:
                    role = Role.get_on_index(uuid, 'slug')
                except KeyError:
                    try:
                        role = Role.get_on_index(slug, 'slug')
                    except KeyError:
                        # New role
                        if action != 'provision':
                            continue
                        role = Role()
                        role.id = uuid
            if action == 'provision':
                # Provision/rename
                role.name = name
                role.slug = uuid
                role.emails = emails
                role.details = details
                role.emails_to_members = emails_to_members
                role.store()
            elif action == 'deprovision':
                # Deprovision
                role.remove_self()
        # All roles have been sent
        if full:
            for role in Role.select():
                if role.slug not in uuids:
                    role.remove_self()

CmdHoboNotify.register()
